package com.chelvc.framework.database.interceptor;

import java.util.Collection;
import java.util.Collections;
import java.util.Date;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.stream.Collectors;

import com.baomidou.mybatisplus.core.conditions.query.LambdaQueryWrapper;
import com.chelvc.framework.base.context.SessionContextHolder;
import com.chelvc.framework.common.exception.ParameterInvalidException;
import com.chelvc.framework.common.util.ObjectUtils;
import com.chelvc.framework.database.context.DatabaseContextHolder;
import com.chelvc.framework.database.context.UniqueContext;
import com.chelvc.framework.database.entity.Creatable;
import com.chelvc.framework.database.entity.Deletable;
import com.chelvc.framework.database.entity.Entity;
import com.chelvc.framework.database.entity.Updatable;
import lombok.extern.slf4j.Slf4j;
import org.apache.ibatis.binding.MapperMethod;
import org.apache.ibatis.cache.CacheKey;
import org.apache.ibatis.executor.Executor;
import org.apache.ibatis.mapping.BoundSql;
import org.apache.ibatis.mapping.MappedStatement;
import org.apache.ibatis.mapping.SqlCommandType;
import org.apache.ibatis.plugin.Interceptor;
import org.apache.ibatis.plugin.Intercepts;
import org.apache.ibatis.plugin.Invocation;
import org.apache.ibatis.plugin.Plugin;
import org.apache.ibatis.plugin.Signature;
import org.apache.ibatis.session.ResultHandler;
import org.apache.ibatis.session.RowBounds;
import org.springframework.stereotype.Component;

/**
 * 基础属性自动更新维护拦截器
 *
 * @author Woody
 * @date 2024/1/30
 */
@Slf4j
@Component
@Intercepts({
        @Signature(type = Executor.class, method = "update", args = {MappedStatement.class, Object.class}),
        @Signature(type = Executor.class, method = "query", args = {MappedStatement.class, Object.class,
                RowBounds.class, ResultHandler.class}),
        @Signature(type = Executor.class, method = "query", args = {MappedStatement.class, Object.class,
                RowBounds.class, ResultHandler.class, CacheKey.class, BoundSql.class})
})
public class PropertyUpdateInterceptor implements Interceptor {
    /**
     * 初始化实体默认属性值
     *
     * @param entity 数据实体
     */
    private void initializeDefaultValue(Entity<?> entity) {
        Date now = new Date();
        Long operator = SessionContextHolder.getId();
        if (entity instanceof Creatable) {
            Creatable<?> creatable = (Creatable<?>) entity;
            if (creatable.getCreator() == null) {
                creatable.setCreator(operator);
            }
            if (creatable.getCreateTime() == null) {
                creatable.setCreateTime(now);
            }
        }
        if (entity instanceof Updatable) {
            Updatable<?> updatable = (Updatable<?>) entity;
            updatable.setUpdater(operator);
            updatable.setUpdateTime(now);
        }
        if (entity instanceof Deletable) {
            Deletable<?> deletable = (Deletable<?>) entity;
            if (deletable.getDeleted() == null) {
                deletable.setDeleted(false);
            }
        }
    }

    /**
     * 数据查询
     *
     * @param invocation 方法调用信息
     * @return 查询结果
     * @throws Throwable 操作异常
     */
    private Object select(Invocation invocation) throws Throwable {
        // 加密数据解码
        Object result = invocation.proceed();
        if (ObjectUtils.notBlank(result) && DatabaseContextHolder.isCodecEnabled()) {
            try {
                DatabaseContextHolder.decoder(cipher -> ObjectUtils.stream(result).forEach(
                        entity -> DatabaseContextHolder.decode(cipher, entity)
                ));
            } catch (Exception e) {
                log.warn("Confidential value decode failed: {}", e.getMessage());
            }
        }
        return result;
    }

    /**
     * 数据更新
     *
     * @param invocation 方法调用信息
     * @return 更新结果
     * @throws Throwable 操作异常
     */
    private Object modify(Invocation invocation) throws Throwable {
        // 获取更新目标对象实例
        Object object = invocation.getArgs()[1];
        if (object instanceof MapperMethod.ParamMap) {
            // 如果是批量更新则为ParamMap
            // 如果Mapper方法使用@Param参数注解则map包含"param1"和自定义参数名称，否则使用"collection"获取列表
            // ParamMap如果get不存在的key会排除异常，所以先判断目标key是否存在
            MapperMethod.ParamMap<?> param = (MapperMethod.ParamMap<?>) object;
            if (param.containsKey("collection")) {
                object = param.get("collection");
            } else if (param.containsKey("param1")) {
                object = param.get("param1");
            }
        }

        // 数据更新
        if (object instanceof Collection) {
            ((Collection<?>) object).forEach(entity -> {
                if (entity instanceof Entity) {
                    this.modify((Entity<?>) entity);
                }
            });
        } else if (object instanceof Entity) {
            this.modify((Entity<?>) object);
        }
        return invocation.proceed();
    }

    /**
     * 实体信息更新
     *
     * @param entity 实体对象实例
     * @param <T>    实体类型
     */
    @SuppressWarnings("unchecked")
    private <T extends Entity<?>> void modify(T entity) {
        // 初始化实体默认属性值
        this.initializeDefaultValue(entity);

        // 设置加密属性值
        Class<T> clazz = (Class<T>) entity.getClass();
        boolean encoding = DatabaseContextHolder.isCodecEnabled() && DatabaseContextHolder.isConfidential(clazz);
        T original = encoding ? ObjectUtils.copying(entity, clazz) : entity;
        if (encoding) {
            DatabaseContextHolder.encode(entity);
        }

        // 唯一属性值校验
        // 根据唯一属性值查询现有记录
        List<UniqueContext<T>> uniques = DatabaseContextHolder.getUniqueContexts(clazz);
        if (ObjectUtils.isEmpty(uniques)) {
            return;
        }
        LambdaQueryWrapper<T> query = DatabaseContextHolder.initializeUniqueQuery(clazz, entity, uniques);
        List<T> objects = query.isEmptyOfWhere() ? Collections.emptyList() :
                DatabaseContextHolder.lookupEntityService(clazz).list(query);
        if (ObjectUtils.isEmpty(objects)) {
            return;
        }

        // 排除ID相同的数据
        Object id = DatabaseContextHolder.getEntityIdentity(entity);
        objects.removeIf(object -> Objects.equals(id, DatabaseContextHolder.getEntityIdentity(object)));

        // 比较唯一属性值，找出已存在唯一属性名称
        Map<String, String> errors = uniques.stream().filter(context -> objects.stream().anyMatch(
                object -> context.getGetters().stream().allMatch(getter -> {
                    Object value = getter.apply(object);
                    if (encoding) {
                        return Objects.equals(value, getter.apply(original))
                                || Objects.equals(value, getter.apply(entity));
                    }
                    return Objects.equals(value, getter.apply(original));
                })
        )).collect(Collectors.toMap(
                context -> context.getField().getName(), context -> context.getUnique().message()
        ));
        if (ObjectUtils.notEmpty(errors)) {
            throw new ParameterInvalidException(errors);
        }
    }

    @Override
    public Object intercept(Invocation invocation) throws Throwable {
        if (!(invocation.getTarget() instanceof Executor)) {
            return invocation.proceed();
        }
        SqlCommandType operation = ((MappedStatement) invocation.getArgs()[0]).getSqlCommandType();
        if (operation == SqlCommandType.SELECT) {
            return this.select(invocation);
        } else if (operation == SqlCommandType.INSERT || operation == SqlCommandType.UPDATE) {
            return this.modify(invocation);
        }
        return invocation.proceed();
    }

    @Override
    public Object plugin(Object target) {
        return target instanceof Executor ? Plugin.wrap(target, this) : target;
    }
}
