package com.chelvc.framework.export.support;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.io.Serializable;
import java.time.Duration;
import java.util.Collection;
import java.util.List;
import java.util.Objects;
import java.util.function.BiConsumer;
import java.util.function.BiFunction;
import java.util.stream.Collectors;
import javax.servlet.http.HttpServletResponse;

import com.chelvc.framework.base.context.ApplicationContextHolder;
import com.chelvc.framework.base.context.ThreadContextHolder;
import com.chelvc.framework.base.util.HttpUtils;
import com.chelvc.framework.common.function.Executor;
import com.chelvc.framework.common.util.AssertUtils;
import com.chelvc.framework.common.util.ExcelUtils;
import com.chelvc.framework.common.util.FileUtils;
import com.chelvc.framework.common.util.ObjectUtils;
import com.chelvc.framework.common.util.ThreadUtils;
import com.chelvc.framework.export.ExportHandler;
import com.chelvc.framework.export.Exporting;
import com.chelvc.framework.export.config.ExportProperties;
import com.chelvc.framework.redis.context.RedisContextHolder;
import com.google.common.collect.Lists;
import lombok.NonNull;
import lombok.extern.slf4j.Slf4j;
import org.apache.poi.ss.usermodel.Workbook;
import org.apache.poi.xssf.streaming.SXSSFWorkbook;
import org.springframework.boot.ApplicationArguments;
import org.springframework.boot.ApplicationRunner;
import org.springframework.data.redis.core.RedisTemplate;

/**
 * 数据导出默认实现
 *
 * @author Woody
 * @date 2024/5/9
 */
@Slf4j
public class DefaultExportHandler implements ExportHandler, ApplicationRunner {
    private final String directory;
    private final Duration duration;
    private final RedisTemplate<String, Object> redisTemplate;

    public DefaultExportHandler(@NonNull RedisTemplate<String, Object> redisTemplate,
                                @NonNull ExportProperties properties) {
        this.directory = properties.getDirectory();
        this.duration = Duration.ofSeconds(properties.getExpiration());
        this.redisTemplate = redisTemplate;
    }

    /**
     * 清理导出文件
     */
    private void clear() {
        int batch = 100;
        List<File> files = Lists.newArrayListWithCapacity(batch);
        FileUtils.files(this.directory, file -> {
            files.add(file);
            if (files.size() == batch) {
                this.clear(files);
                files.clear();
            }
        });
        if (ObjectUtils.notEmpty(files)) {
            this.clear(files);
        }
    }

    /**
     * 清理导出文件
     *
     * @param files 文件列表
     */
    private void clear(List<File> files) {
        List<String> keys = files.stream().map(File::getName).map(this::key).collect(Collectors.toList());
        List<?> values = this.redisTemplate.opsForValue().multiGet(keys);
        if (ObjectUtils.notEmpty(values)) {
            for (int i = 0; i < values.size(); i++) {
                if (Objects.isNull(values.get(i))) {
                    FileUtils.delete(files.get(i));
                }
            }
        }
    }

    /**
     * 构建任务标识
     *
     * @param id 任务ID
     * @return 任务标识
     */
    private String key(Serializable id) {
        return "exporting:" + id;
    }

    /**
     * 获取导出任务信息
     *
     * @param id 导出任务ID
     * @return 导出任务信息
     */
    private Exporting exporting(Serializable id) {
        return (Exporting) this.redisTemplate.opsForValue().get(this.key(id));
    }

    /**
     * 设置导出任务信息
     *
     * @param id       导出任务ID
     * @param filename 文件名称
     * @param status   导出状态
     * @param duration 任务有效期
     */
    private void exporting(Serializable id, String filename, Exporting.Status status, Duration duration) {
        Exporting exporting = Exporting.builder().filename(filename).status(status).build();
        this.redisTemplate.opsForValue().set(this.key(id), exporting, duration);
    }

    @Override
    public void run(ApplicationArguments args) throws Exception {
        // 初始化导出文件清理线程
        ThreadUtils.run(() -> {
            log.info("Export file cleaner running...");
            while (!Thread.currentThread().isInterrupted()) {
                try {
                    RedisContextHolder.tryLockAround("exporting:clear:lock", (Executor) this::clear);
                } catch (Exception e) {
                    log.error("Export file clear failed", e);
                }
                ThreadUtils.sleep(this.duration.toMillis());
            }
        });
    }

    @Override
    public Exporting.Status status(long id) {
        return ObjectUtils.ifNull(this.exporting(id), Exporting::getStatus);
    }

    @Override
    public void download(long id, @NonNull HttpServletResponse response) throws IOException {
        Exporting exporting = this.exporting(id);
        AssertUtils.available(exporting, () -> ApplicationContextHolder.getMessage("Export.Expired"));
        boolean completed = ObjectUtils.ifNull(exporting, Exporting::getStatus) == Exporting.Status.COMPLETED;
        AssertUtils.available(completed, () -> ApplicationContextHolder.getMessage("Export.Uncompleted"));
        String filename = ObjectUtils.ifNull(exporting, Exporting::getFilename);
        HttpUtils.write(response, new File(this.directory, String.valueOf(id)), filename);
    }

    @Override
    public <T> long export(@NonNull String filename, @NonNull BiConsumer<Long, Integer> listener,
                           @NonNull BiFunction<Integer, Integer, Collection<T>> provider,
                           @NonNull ExcelUtils.Writer<T> writer, @NonNull String... titles) {
        // 生成导出任务ID
        long id = RedisContextHolder.identity();

        // 初始化数据导出状态
        this.exporting(id, filename, Exporting.Status.RUNNING, Duration.ofDays(1));

        // 异步执行导出任务
        ThreadContextHolder.execute(() -> {
            int volume;
            try {
                // 将数据写入工作薄
                Workbook workbook = new SXSSFWorkbook();
                volume = ExcelUtils.write(workbook, provider, writer, titles);

                // 保存导出文件
                File file = new File(this.directory, String.valueOf(id));
                try (OutputStream output = new FileOutputStream(file)) {
                    workbook.write(output);
                    workbook.close();
                }

                // 更新导出任务完成状态
                this.exporting(id, filename, Exporting.Status.COMPLETED, this.duration);
            } catch (Exception e) {
                log.error("Export execute failed", e);

                // 更新导出任务失败状态
                this.exporting(id, filename, Exporting.Status.FAILED, this.duration);
                return;
            }

            // 数据导出完成监听回调
            try {
                listener.accept(id, volume);
            } catch (Exception e) {
                log.error("Export listening failed", e);
            }
        });
        return id;
    }
}
