package com.chelvc.framework.oauth.token;

import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import com.chelvc.framework.base.context.SessionContextHolder;
import com.chelvc.framework.common.model.Terminal;
import com.chelvc.framework.common.util.ObjectUtils;
import com.chelvc.framework.common.util.StringUtils;
import com.chelvc.framework.oauth.context.OAuthContextHolder;
import com.chelvc.framework.redis.context.RedisContextHolder;
import lombok.NonNull;
import lombok.extern.slf4j.Slf4j;
import org.springframework.data.redis.core.script.DefaultRedisScript;
import org.springframework.data.redis.core.script.RedisScript;
import org.springframework.security.core.userdetails.UserDetails;
import org.springframework.security.oauth2.common.OAuth2AccessToken;
import org.springframework.security.oauth2.provider.OAuth2Authentication;
import org.springframework.security.oauth2.provider.token.store.JwtAccessTokenConverter;
import org.springframework.security.oauth2.provider.token.store.JwtTokenStore;

/**
 * 基于Redis令牌管理器实现
 *
 * @author Woody
 * @date 2024/1/30
 */
@Slf4j
public class RedisTokenStore extends JwtTokenStore {
    /**
     * 终端/令牌更新脚本映射表
     */
    private final Map<Terminal, RedisScript<Boolean>> scripts;

    public RedisTokenStore(@NonNull JwtAccessTokenConverter jwtAccessTokenConverter) {
        super(jwtAccessTokenConverter);

        // 初始化各终端令牌更新脚本映射表
        this.scripts = Stream.of(Terminal.values()).collect(Collectors.toMap(terminal -> terminal,
                terminal -> new DefaultRedisScript<>(
                        String.format("return redis.call('HSET', KEYS[1], 'scope', ARGV[1], '%s', ARGV[2]) " +
                                "and redis.call('EXPIRE', KEYS[1], ARGV[3])", terminal), Boolean.class
                )));

        // 初始化无终端令牌更新脚本
        this.scripts.put(null, new DefaultRedisScript<>(
                "return redis.call('HSET', KEYS[1], 'scope', ARGV[1], 'null', ARGV[2]) " +
                        "and redis.call('EXPIRE', KEYS[1], ARGV[3])", Boolean.class
        ));
    }

    @Override
    public void storeAccessToken(OAuth2AccessToken token, OAuth2Authentication authentication) {
        Object principal = authentication.getPrincipal();
        if (principal instanceof UserDetails) {
            principal = ((UserDetails) principal).getUsername();
        }
        List<String> keys = Collections.singletonList(OAuthContextHolder.key(principal));
        String scope = ObjectUtils.ifNull(OAuthContextHolder.getScope(token), StringUtils.EMPTY);
        long duration = RedisContextHolder.duration(token.getExpiration()).getSeconds();
        RedisScript<Boolean> script = this.scripts.get(SessionContextHolder.getTerminal());
        try {
            RedisContextHolder.getDefaultTemplate().execute(script, keys, scope, token.getValue(), duration);
        } catch (Exception e) {
            log.warn("Redis token save failed: {}", e.getMessage());
        }
    }
}
